/*----------------------------------------------------------------------------
 *
 * similarity.c
 *
 * Copyright (c) 2008-2009, Euler Taveira de Oliveira
 *
 *----------------------------------------------------------------------------
 */

#include "similarity.h"

PG_MODULE_MAGIC;

/*
 * Monge-Elkan approximate sets
 */
static char *approx_set[7] = {
	"dt",
	"gj",
	"lr",
	"mn",
	"bpv",
	"aeiou",
	",."
};

/*
 * cost functions
 */
int levcost(char a, char b)
{
	if (a == b)
		return PGS_LEV_MIN_COST;
	else
		return PGS_LEV_MAX_COST;
}

/*
 * TODO change it to a callback function
 */
int nwcost(char a, char b)
{
	if (a == 'a' && b == 'a')
		return 10;
	else if (a == 'a' && b == 'g')
		return -1;
	else if (a == 'a' && b == 'c')
		return -3;
	else if (a == 'a' && b == 't')
		return -4;
	else if (a == 'g' && b == 'a')
		return -1;
	else if (a == 'g' && b == 'g')
		return 7;
	else if (a == 'g' && b == 'c')
		return -5;
	else if (a == 'g' && b == 't')
		return -3;
	else if (a == 'c' && b == 'a')
		return -3;
	else if (a == 'c' && b == 'g')
		return -5;
	else if (a == 'c' && b == 'c')
		return 9;
	else if (a == 'c' && b == 't')
		return 0;
	else if (a == 't' && b == 'a')
		return -4;
	else if (a == 't' && b == 'g')
		return -3;
	else if (a == 't' && b == 'c')
		return 0;
	else if (a == 't' && b == 't')
		return 8;
	else
		return -99;	/* shouldn't happen */
}

float swcost(char *a, char *b, int i, int j)
{
	/* XXX paranoia? check for out-of-range index */
	if (i < 0 || i >= strlen(a))
		return 0.0;
	if (j < 0 || j >= strlen(b))
		return 0.0;

	if (a[i] == b[j])
		return PGS_SW_MAX_COST;
	else
		return PGS_SW_MIN_COST;
}

float swggapcost(int i, int j)
{
	if (i >= j)
		return 0.0;
	else
		return (5.0 + ((j - 1)  - i));
}

float megapcost(char *a, char *b, int i, int j)
{
	int k;

	/* XXX paranoia? check for out-of-range index */
	if (i < 0 || i >= strlen(a))
		return -3.0;
	if (j < 0 || j >= strlen(b))
		return -3.0;

	if (a[i] == b[j])
		return 5.0;

	for (k = 0; k < 7; k++)
	{
		if (strchr(approx_set[k], a[i]) != NULL &&
			strchr(approx_set[k], b[j]) != NULL)
			return 3.0;
	}

	return -3.0;
}

/*
 * get/set isNormalized function
 */
PG_FUNCTION_INFO_V1(set_isnormalized);

Datum
set_isnormalized(PG_FUNCTION_ARGS)
{
	char	*f = DatumGetPointer(DirectFunctionCall1(textout, PointerGetDatum(PG_GETARG_TEXT_P(0))));
	bool	t = PG_GETARG_BOOL(1);

	if (strcmp(f, "block") == 0)
		block_is_normalized = t;
	else if (strcmp(f , "cosine") == 0)
		cosine_is_normalized = t;
	else if (strcmp(f , "dice") == 0)
		dice_is_normalized = t;
	else if (strcmp(f , "euclidean") == 0)
		euclidean_is_normalized = t;
	else if (strcmp(f , "hamming") == 0)
		hamming_is_normalized = t;
	else if (strcmp(f , "jaccard") == 0)
		jaccard_is_normalized = t;
	else if (strcmp(f , "jaro") == 0)
		jaro_is_normalized = t;
	else if (strcmp(f , "jarowinkler") == 0)
		jarowinkler_is_normalized = t;
	else if (strcmp(f , "levenshtein") == 0)
		levenshtein_is_normalized = t;
	else if (strcmp(f , "matching") == 0)
		matching_is_normalized = t;
	else if (strcmp(f , "mongeelkan") == 0)
		mongeelkan_is_normalized = t;
	else if (strcmp(f , "needlemanwunsch") == 0)
		needlemanwunsch_is_normalized = t;
	else if (strcmp(f , "overlapcoefficient") == 0)
		overlap_is_normalized = t;
	else if (strcmp(f , "qgram") == 0)
		qgram_is_normalized = t;
	else if (strcmp(f , "smithwaterman") == 0)
		smithwaterman_is_normalized = t;
	else if (strcmp(f , "smithwatermangotoh") == 0)
		smithwatermangotoh_is_normalized = t;
	else
		elog(ERROR, "function \"%s\" doesn't exist", f);

	PG_RETURN_BOOL(t);
}

PG_FUNCTION_INFO_V1(get_isnormalized);

Datum
get_isnormalized(PG_FUNCTION_ARGS)
{
	char	*f = DatumGetPointer(DirectFunctionCall1(textout, PointerGetDatum(PG_GETARG_TEXT_P(0))));
	bool	t;

	if (strcmp(f, "block") == 0)
		t = block_is_normalized;
	else if (strcmp(f , "cosine") == 0)
		t = cosine_is_normalized;
	else if (strcmp(f , "dice") == 0)
		t = dice_is_normalized;
	else if (strcmp(f , "euclidean") == 0)
		t = euclidean_is_normalized;
	else if (strcmp(f , "hamming") == 0)
		t = hamming_is_normalized;
	else if (strcmp(f , "jaccard") == 0)
		t = jaccard_is_normalized;
	else if (strcmp(f , "jaro") == 0)
		t = jaro_is_normalized;
	else if (strcmp(f , "jarowinkler") == 0)
		t = jarowinkler_is_normalized;
	else if (strcmp(f , "levenshtein") == 0)
		t = levenshtein_is_normalized;
	else if (strcmp(f , "matching") == 0)
		t = matching_is_normalized;
	else if (strcmp(f , "mongeelkan") == 0)
		t = mongeelkan_is_normalized;
	else if (strcmp(f , "needlemanwunsch") == 0)
		t = needlemanwunsch_is_normalized;
	else if (strcmp(f , "overlapcoefficient") == 0)
		t = overlap_is_normalized;
	else if (strcmp(f , "qgram") == 0)
		t = qgram_is_normalized;
	else if (strcmp(f , "smithwaterman") == 0)
		t = smithwaterman_is_normalized;
	else if (strcmp(f , "smithwatermangotoh") == 0)
		t = smithwatermangotoh_is_normalized;
	else
		elog(ERROR, "function \"%s\" doesn't exist", f);

	PG_RETURN_BOOL(t);
}

/*
 * get/set threshold function
 */
PG_FUNCTION_INFO_V1(set_threshold);

Datum
set_threshold(PG_FUNCTION_ARGS)
{
	char	*f = DatumGetPointer(DirectFunctionCall1(textout, PointerGetDatum(PG_GETARG_TEXT_P(0))));
	float4	t = PG_GETARG_FLOAT4(1);

	if (t < 0.0 || t > 1.0)
		elog(ERROR, "threshold must be between 0.0 and 1.0");

	if (strcmp(f, "block") == 0)
		block_threshold = t;
	else if (strcmp(f , "cosine") == 0)
		cosine_threshold = t;
	else if (strcmp(f , "dice") == 0)
		dice_threshold = t;
	else if (strcmp(f , "euclidean") == 0)
		euclidean_threshold = t;
	else if (strcmp(f , "hamming") == 0)
		hamming_threshold = t;
	else if (strcmp(f , "jaccard") == 0)
		jaccard_threshold = t;
	else if (strcmp(f , "jaro") == 0)
		jaro_threshold = t;
	else if (strcmp(f , "jarowinkler") == 0)
		jarowinkler_threshold = t;
	else if (strcmp(f , "levenshtein") == 0)
		levenshtein_threshold = t;
	else if (strcmp(f , "matching") == 0)
		matching_threshold = t;
	else if (strcmp(f , "mongeelkan") == 0)
		mongeelkan_threshold = t;
	else if (strcmp(f , "needlemanwunsch") == 0)
		needlemanwunsch_threshold = t;
	else if (strcmp(f , "overlapcoefficient") == 0)
		overlap_threshold = t;
	else if (strcmp(f , "qgram") == 0)
		qgram_threshold = t;
	else if (strcmp(f , "smithwaterman") == 0)
		smithwaterman_threshold = t;
	else if (strcmp(f , "smithwatermangotoh") == 0)
		smithwatermangotoh_threshold = t;
	else
		elog(ERROR, "function \"%s\" doesn't exist", f);

	PG_RETURN_FLOAT4(t);
}

PG_FUNCTION_INFO_V1(get_threshold);

Datum
get_threshold(PG_FUNCTION_ARGS)
{
	char	*f = DatumGetPointer(DirectFunctionCall1(textout, PointerGetDatum(PG_GETARG_TEXT_P(0))));
	float4	t;

	if (strcmp(f, "block") == 0)
		t = block_threshold;
	else if (strcmp(f , "cosine") == 0)
		t = cosine_threshold;
	else if (strcmp(f , "dice") == 0)
		t = dice_threshold;
	else if (strcmp(f , "euclidean") == 0)
		t = euclidean_threshold;
	else if (strcmp(f , "hamming") == 0)
		t = hamming_threshold;
	else if (strcmp(f , "jaccard") == 0)
		t = jaccard_threshold;
	else if (strcmp(f , "jaro") == 0)
		t = jaro_threshold;
	else if (strcmp(f , "jarowinkler") == 0)
		t = jarowinkler_threshold;
	else if (strcmp(f , "levenshtein") == 0)
		t = levenshtein_threshold;
	else if (strcmp(f , "matching") == 0)
		t = matching_threshold;
	else if (strcmp(f , "mongeelkan") == 0)
		t = mongeelkan_threshold;
	else if (strcmp(f , "needlemanwunsch") == 0)
		t = needlemanwunsch_threshold;
	else if (strcmp(f , "overlapcoefficient") == 0)
		t = overlap_threshold;
	else if (strcmp(f , "qgram") == 0)
		t = qgram_threshold;
	else if (strcmp(f , "smithwaterman") == 0)
		t = smithwaterman_threshold;
	else if (strcmp(f , "smithwatermangotoh") == 0)
		t = smithwatermangotoh_threshold;
	else
		elog(ERROR, "function \"%s\" doesn't exist", f);

	PG_RETURN_FLOAT4(t);
}

/*
 * get/set tokenizer function
 */
PG_FUNCTION_INFO_V1(set_tokenizer);

Datum
set_tokenizer(PG_FUNCTION_ARGS)
{
	char	*f = DatumGetPointer(DirectFunctionCall1(textout, PointerGetDatum(PG_GETARG_TEXT_P(0))));
	int32	t = PG_GETARG_INT32(1);

	if (t != PGS_UNIT_WORD && t != PGS_UNIT_GRAM && t != PGS_UNIT_ALNUM)
		elog(ERROR, "tokenizer must be %d, %d or %d", PGS_UNIT_WORD, PGS_UNIT_GRAM, PGS_UNIT_ALNUM);


	/*
	 * we're testing functions that don't use a tokenizer to
	 * output a error message that makes sense
	 */
	if (strcmp(f, "block") == 0)
		block_tokenizer = t;
	else if (strcmp(f , "cosine") == 0)
		cosine_tokenizer = t;
	else if (strcmp(f , "dice") == 0)
		dice_tokenizer = t;
	else if (strcmp(f , "euclidean") == 0)
		euclidean_tokenizer = t;
	else if (strcmp(f , "hamming") == 0)
		elog(ERROR, "function \"%s\" doesn't use a tokenizer", f);
	else if (strcmp(f , "jaccard") == 0)
		jaccard_tokenizer = t;
	else if (strcmp(f , "jaro") == 0)
		elog(ERROR, "function \"%s\" doesn't use a tokenizer", f);
	else if (strcmp(f , "jarowinkler") == 0)
		elog(ERROR, "function \"%s\" doesn't use a tokenizer", f);
	else if (strcmp(f , "levenshtein") == 0)
		elog(ERROR, "function \"%s\" doesn't use a tokenizer", f);
	else if (strcmp(f , "matching") == 0)
		matching_tokenizer = t;
	else if (strcmp(f , "mongeelkan") == 0)
		mongeelkan_tokenizer = t;
	else if (strcmp(f , "needlemanwunsch") == 0)
		elog(ERROR, "function \"%s\" doesn't use a tokenizer", f);
	else if (strcmp(f , "overlapcoefficient") == 0)
		overlap_tokenizer = t;
	else if (strcmp(f , "qgram") == 0)
		qgram_tokenizer = t;
	else if (strcmp(f , "smithwaterman") == 0)
		elog(ERROR, "function \"%s\" doesn't use a tokenizer", f);
	else if (strcmp(f , "smithwatermangotoh") == 0)
		elog(ERROR, "function \"%s\" doesn't use a tokenizer", f);
	else
		elog(ERROR, "function \"%s\" doesn't exist", f);

	PG_RETURN_INT32(t);
}

PG_FUNCTION_INFO_V1(get_tokenizer);

Datum
get_tokenizer(PG_FUNCTION_ARGS)
{
	char	*f = DatumGetPointer(DirectFunctionCall1(textout, PointerGetDatum(PG_GETARG_TEXT_P(0))));
	int32	t;

	/*
	 * we're testing functions that don't use a tokenizer to
	 * output a error message that makes sense
	 */
	if (strcmp(f, "block") == 0)
		t = block_tokenizer;
	else if (strcmp(f , "cosine") == 0)
		t = cosine_tokenizer;
	else if (strcmp(f , "dice") == 0)
		t = dice_tokenizer;
	else if (strcmp(f , "euclidean") == 0)
		t = euclidean_tokenizer;
	else if (strcmp(f , "hamming") == 0)
		elog(ERROR, "function \"%s\" doesn't use a tokenizer", f);
	else if (strcmp(f , "jaccard") == 0)
		t = jaccard_tokenizer;
	else if (strcmp(f , "jaro") == 0)
		elog(ERROR, "function \"%s\" doesn't use a tokenizer", f);
	else if (strcmp(f , "jarowinkler") == 0)
		elog(ERROR, "function \"%s\" doesn't use a tokenizer", f);
	else if (strcmp(f , "levenshtein") == 0)
		elog(ERROR, "function \"%s\" doesn't use a tokenizer", f);
	else if (strcmp(f , "matching") == 0)
		t = matching_tokenizer;
	else if (strcmp(f , "mongeelkan") == 0)
		t = mongeelkan_tokenizer;
	else if (strcmp(f , "needlemanwunsch") == 0)
		elog(ERROR, "function \"%s\" doesn't use a tokenizer", f);
	else if (strcmp(f , "overlapcoefficient") == 0)
		t = overlap_tokenizer;
	else if (strcmp(f , "qgram") == 0)
		t = qgram_tokenizer;
	else if (strcmp(f , "smithwaterman") == 0)
		elog(ERROR, "function \"%s\" doesn't use a tokenizer", f);
	else if (strcmp(f , "smithwatermangotoh") == 0)
		elog(ERROR, "function \"%s\" doesn't use a tokenizer", f);
	else
		elog(ERROR, "function \"%s\" doesn't exist", f);

	PG_RETURN_INT32(t);
}
